@@ -1,4 +1,4 @@
-// Copyright 2016, Google
+// Copyright 2016, the Blazer authors
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -71,7 +71,7 @@ type testRoot struct {
 	bucketMap map[string]map[string]string
 }
 
-func (t *testRoot) authorizeAccount(context.Context, string, string, ...ClientOption) error {
+func (t *testRoot) authorizeAccount(context.Context, string, string, clientOptions) error {
 	t.auths++
 	return nil
 }
@@ -108,6 +108,13 @@ func (t *testRoot) transient(err error) bool {
 	return e.retry || e.reupload || e.backoff > 0
 }
 
+func (t *testRoot) createKey(context.Context, string, []string, time.Duration, string, string) (b2KeyInterface, error) {
+	return nil, nil
+}
+func (t *testRoot) listKeys(context.Context, int, string) ([]b2KeyInterface, string, error) {
+	return nil, "", nil
+}
+
 func (t *testRoot) createBucket(_ context.Context, name, _ string, _ map[string]string, _ []LifecycleRule) (b2BucketInterface, error) {
 	if err := t.errs.getError("createBucket"); err != nil {
 		return nil, err
@@ -147,6 +154,7 @@ func (t *testBucket) btype() string                                    { return
 func (t *testBucket) attrs() *BucketAttrs                              { return nil }
 func (t *testBucket) deleteBucket(context.Context) error               { return nil }
 func (t *testBucket) updateBucket(context.Context, *BucketAttrs) error { return nil }
+func (t *testBucket) id() string                                       { return "" }
 
 func (t *testBucket) getUploadURL(context.Context) (b2URLInterface, error) {
 	if err := t.errs.getError("getUploadURL"); err != nil {
@@ -221,7 +229,7 @@ func (t *testBucket) downloadFileByName(_ context.Context, name string, offset,
 }
 
 func (t *testBucket) hideFile(context.Context, string) (b2FileInterface, error) { return nil, nil }
-func (t *testBucket) getDownloadAuthorization(context.Context, string, time.Duration) (string, error) {
+func (t *testBucket) getDownloadAuthorization(context.Context, string, time.Duration, string) (string, error) {
 	return "", nil
 }
 func (t *testBucket) baseURL() string                      { return "" }
@@ -279,6 +287,8 @@ func (t *testLargeFile) getUploadPartURL(context.Context) (b2FileChunkInterface,
 	}, nil
 }
 
+func (t *testLargeFile) cancel(ctx context.Context) error { return ctx.Err() }
+
 type testFileChunk struct {
 	parts map[int][]byte
 	errs  *errCont
@@ -672,6 +682,50 @@ func TestReadWrite(t *testing.T) {
 	}
 }
 
+func TestLargeFileCancellation(t *testing.T) {
+	ctx := context.Background()
+	ctx, cancel := context.WithCancel(ctx)
+	defer cancel()
+
+	client := &Client{
+		backend: &beRoot{
+			b2i: &testRoot{
+				bucketMap: make(map[string]map[string]string),
+				errs:      &errCont{},
+			},
+		},
+	}
+
+	b, err := client.NewBucket(ctx, bucketName, nil)
+	if err != nil {
+		t.Fatal(err)
+	}
+	var called bool
+	w := b.Object("foo").NewWriter(ctx, WithCancelOnError(func() context.Context { return context.Background() }, func(err error) {
+		called = true
+		if err != nil {
+			t.Errorf("expected no error, got %v", err)
+		}
+	}))
+	w.ChunkSize = 10
+
+	for i := 0; i < 10; i++ {
+		r := io.LimitReader(zReader{}, 20)
+		if _, err := io.Copy(w, r); err != nil && err != context.Canceled {
+			t.Errorf("Copy: %v", err)
+		}
+		cancel()
+	}
+
+	if err := w.Close(); err != context.Canceled {
+		t.Errorf("expected cancelled context; got %v", err)
+	}
+
+	if !called {
+		t.Errorf("error callback not called")
+	}
+}
+
 func TestReadRangeReturnsRight(t *testing.T) {
 	ctx := context.Background()
 	ctx, cancel := context.WithTimeout(ctx, 10*time.Second)
