@@ -6,7 +6,9 @@ import (
 	"os"
 	"strconv"
 	"sync"
+	"sync/atomic"
 	"testing"
+	"time"
 
 	"github.com/kurin/blazer/b2"
 )
@@ -30,9 +32,9 @@ func TestOperationLive(t *testing.T) {
 		wg.Add(1)
 		i := i
 		go func() {
-			var n int
 			defer wg.Done()
 			for j := 0; j < 10; j++ {
+				var n int
 				if err := g.Operate(ctx, name, func(b []byte) ([]byte, error) {
 					if len(b) > 0 {
 						i, err := strconv.Atoi(string(b))
@@ -123,6 +125,38 @@ func TestOperationJSONLive(t *testing.T) {
 	}
 }
 
+func TestMutex(t *testing.T) {
+	ctx := context.Background()
+	bucket, done := startLiveTest(ctx, t)
+	defer done()
+
+	g := NewGroup(bucket, "tester")
+	m := g.Mutex(ctx, "mootex")
+	var a int32
+	var wg sync.WaitGroup
+	for i := 0; i < 5; i++ {
+		wg.Add(1)
+		go func(i int) {
+			defer wg.Done()
+			for j := 0; j < 5; j++ {
+				m.Lock()
+				new := atomic.AddInt32(&a, 1)
+				if new != 1 {
+					t.Fatalf("two threads locked at once")
+				}
+				time.Sleep(20 * time.Millisecond)
+				new = atomic.AddInt32(&a, -1)
+				if new != 0 {
+					t.Fatalf("two threads locked at once")
+				}
+				t.Logf("thread %d: lock %d", i, j)
+				m.Unlock()
+			}
+		}(i)
+	}
+	wg.Wait()
+}
+
 func startLiveTest(ctx context.Context, t *testing.T) (*b2.Bucket, func()) {
 	id := os.Getenv(apiID)
 	key := os.Getenv(apiKey)
